deferred class
	CE_ARRAY [G]

inherit
	CE_POINTER [G]
		rename
			make as make_from_pointer,
			make_shared as make_shared_from_pointer,
			make_shared_with_count as make_shared_with_count_from_pointer,
			make_cast as make_cast_from_pointer
		redefine
			make_from_pointer,
			make_shared_from_pointer,
			make_shared_with_count_from_pointer,
			make_cast_from_pointer,
			copy,
			is_equal
		end
	DISPOSABLE
		rename
			is_equal as is_equal_from_any
		undefine
			is_equal_from_any
		redefine
			copy
		end

feature {CE_ARRAY, CE_CLASS}

	make (a_dimensions: ARRAY [INTEGER_32])
		require
			a_dimensions /= Void
			a_dimensions.count = 1
			a_dimensions[1] >= 0
		do
			byte_count := a_dimensions[1] * element_size
			c_pointer := c_pointer.memory_calloc (a_dimensions[1], element_size)
			total_count := a_dimensions[1]
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER; a_dimensions: ARRAY [INTEGER_32])
		require
			a_dimensions /= Void
			a_dimensions.count = 1
			a_dimensions[1] >= 0
		do
			byte_count := a_dimensions[1] * element_size
			c_pointer := a_c_pointer
			total_count := a_dimensions[1]
		end

	make_shared_from_pointer (a_c_pointer: POINTER)
		do
			byte_count := {INTEGER_32}.Max_value
			c_pointer := a_c_pointer
			total_count := {INTEGER_32}.Max_value
		end

feature {NONE}

	make_initialized (a_data: ARRAY [G])
		require
			a_data /= Void and then not a_data.is_empty
		do
			make (<<a_data.count>>)
			across a_data as c loop
				set_i_th (c.item, c.target_index - 1)
			end
		end

feature {ANY}

	total_count: INTEGER_32
		-- count:       number of elements (only first dimensions if multiple)
		-- total_count: number of all elements (in all dimensions)
		-- byte_count:  count, but in bytes

feature {ANY}

	copy (a_other: like Current)
		do
			make (<<a_other.total_count>>)
			c_pointer.memory_copy (a_other.c_pointer, total_count * element_size)
		end

	is_equal (a_other: CE_POINTER [ANY]): BOOLEAN
		do
			Result :=
				Precursor (a_other) or else
				(
					attached {CE_ARRAY [G]} a_other as tmp and then
					total_count = tmp.total_count and then
					c_pointer.memory_compare (a_other.c_pointer, total_count * element_size)
				)
		end

feature {NONE}

	is_allocated_memory: BOOLEAN

	dispose
		do
			if is_allocated_memory then
				c_pointer.memory_free
			end
		end

feature {NONE}

	make_from_pointer
		do
			(create {EXCEPTIONS}).raise ("Not possible for arrays.")
		end

	make_shared_with_count_from_pointer (a_c_pointer: POINTER; a_count: INTEGER_32)
		do
			(create {EXCEPTIONS}).raise ("Not possible for arrays.")
		end

	make_cast_from_pointer (a_other: CE_POINTER [ANY])
		do
			(create {EXCEPTIONS}).raise ("Not possible for arrays.")
		end

invariant
	c_pointer /= default_pointer

end
