deferred class
	CE_POINTER [G]

inherit
	ANY
		rename
			is_equal as is_equal_from_any
		redefine
			is_equal_from_any
		end

feature {CE_POINTER, CE_CLASS}

	make
		do
			byte_count := {INTEGER_32}.Max_value
		end

	make_shared (a_c_pointer: POINTER)
		do
			byte_count := {INTEGER_32}.Max_value
			c_pointer := a_c_pointer
		end

	make_shared_with_count (a_c_pointer: POINTER; a_count: INTEGER_32)
		do
			byte_count := a_count * element_size
			c_pointer := a_c_pointer
		end

	make_cast (a_other: CE_POINTER [ANY])
		require
			a_other /= Void
		do
			if
				a_other.byte_count /= {INTEGER_32}.Max_value and then
				attached {CE_ARRAY [ANY]} a_other as tmp
			then
				byte_count := tmp.total_count * a_other.element_size
			else
				byte_count := a_other.byte_count
			end
			c_pointer := a_other.c_pointer
		end

feature {ANY}

	c_pointer: POINTER

	byte_count: INTEGER_32

	element_size: INTEGER_32
		deferred
		end

feature {ANY}

	count: INTEGER_32
		do
			-- Don't store the count itself; store the byte count. Otherwise:
			--    int[1] ia; (int[])(long long[])ia
			-- would result in having a count of 0.
			Result := byte_count // element_size
		end

	is_void: BOOLEAN
		do
			Result := c_pointer = default_pointer
		end

	is_equal (a_other: CE_POINTER [ANY]): BOOLEAN
		do
			Result := c_pointer = a_other.c_pointer
		end

	is_equal_from_any (a_other: like Current): BOOLEAN
		do
			Result := attached {CE_POINTER [ANY]} a_other as tmp and then is_equal (tmp)
		end

feature {ANY}

	item: G assign set_item
			-- *Current
		do
			Result := i_th (0)
		end

	set_item (a_item: G)
		do
			set_i_th (a_item, 0)
		end

	i_th alias "[]" (a_index: INTEGER_32): G assign set_i_th
			-- Current[0]
		require
			non_void_target: c_pointer /= default_pointer
			in_bounds: not {CE_SETTINGS}.Disable_bounds_checking implies a_index < count
		deferred
		end

	set_i_th (a_item: G; a_index: INTEGER_32)
			-- Current[0] := foo
		require
			non_void_target: c_pointer /= default_pointer
			in_bounds: not {CE_SETTINGS}.Disable_bounds_checking implies a_index < count
			not_subarray_set: not (attached {CE_ARRAY [CE_ARRAY [ANY]]} Current)
		deferred
		end

feature {ANY} -- pointer arithmetics

	plus alias "+" (i: INTEGER_32): CE_POINTER [G]
		deferred
		end

	minus alias "-" (i: INTEGER_32): CE_POINTER [G]
		do
			Result := plus (-i)
		end

	distance (other: CE_POINTER [G]): INTEGER_32
		do
			Result := (c_pointer.to_integer_32 - other.c_pointer.to_integer_32) // element_size
		end

end
