class
	CE_ROUTINE

inherit
	CE_POINTER [ANY]
		rename
			item as item_from_pointer,
			set_item as set_item_from_pointer,
			i_th as i_th_from_pointer,
			set_i_th as set_i_th_from_pointer,
			plus as plus_from_pointer,
			minus as minus_from_pointer
		redefine
			make,
			make_shared,
			make_cast,
			item_from_pointer,
			set_item_from_pointer,
			i_th_from_pointer,
			set_i_th_from_pointer,
			minus_from_pointer
		end
	INTERNAL
		rename
			is_equal as is_equal_from_any
		undefine
			is_equal_from_any
		end

create
	make,
	make_shared,
	make_agent,
	make_cast

feature {NONE}

	make
		do
		end

	make_shared (a_c_pointer: POINTER)
		do
			c_pointer := a_c_pointer
			routine := agent_c_pointers[a_c_pointer]
		end

	make_agent (a_routine: ROUTINE [ANY, TUPLE])
		require
			a_routine /= Void
		do
			routine := a_routine
			c_pointer := create_c_function_pointer
			agent_c_pointers[c_pointer] := a_routine
		ensure
			routine = a_routine
		end

	make_cast (a_other: CE_POINTER [ANY])
		do
			if
				a_other.byte_count /= {INTEGER_32}.Max_value and then
				attached {CE_ARRAY [ANY]} a_other as tmp
			then
				byte_count := tmp.total_count * a_other.element_size
			else
				byte_count := a_other.byte_count
			end
			c_pointer := a_other.c_pointer
			routine := agent_c_pointers[c_pointer]
		end

feature {CE_ROUTINE}

	routine: detachable ROUTINE [ANY, TUPLE]

feature {ANY}

	call (a_args: TUPLE)
		do
			if attached routine as tmp_routine then
				tmp_routine.call (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_void, default_pointer)
			end
		end

	boolean_item (a_args: TUPLE): BOOLEAN
		do
			if attached {FUNCTION [ANY, TUPLE, BOOLEAN]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_integer_8, $Result)
			end
		end

	integer_8_item (a_args: TUPLE): INTEGER_8
		do
			if attached {FUNCTION [ANY, TUPLE, INTEGER_8]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_integer_8, $Result)
			end
		end

	integer_16_item (a_args: TUPLE): INTEGER_16
		do
			if attached {FUNCTION [ANY, TUPLE, INTEGER_16]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_integer_16, $Result)
			end
		end

	integer_32_item (a_args: TUPLE): INTEGER_32
		do
			if attached {FUNCTION [ANY, TUPLE, INTEGER_32]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_integer_32, $Result)
			end
		end

	integer_64_item (a_args: TUPLE): INTEGER_64
		do
			if attached {FUNCTION [ANY, TUPLE, INTEGER_64]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_integer_64, $Result)
			end
		end

	natural_8_item (a_args: TUPLE): NATURAL_8
		do
			if attached {FUNCTION [ANY, TUPLE, NATURAL_8]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_natural_8, $Result)
			end
		end

	natural_16_item (a_args: TUPLE): NATURAL_16
		do
			if attached {FUNCTION [ANY, TUPLE, NATURAL_16]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_natural_16, $Result)
			end
		end

	natural_32_item (a_args: TUPLE): NATURAL_32
		do
			if attached {FUNCTION [ANY, TUPLE, NATURAL_32]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_natural_32, $Result)
			end
		end

	natural_64_item (a_args: TUPLE): NATURAL_64
		do
			if attached {FUNCTION [ANY, TUPLE, NATURAL_64]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_natural_64, $Result)
			end
		end

	real_32_item (a_args: TUPLE): REAL_32
		do
			if attached {FUNCTION [ANY, TUPLE, REAL_32]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_real_32, $Result)
			end
		end

	real_64_item (a_args: TUPLE): REAL_64
		do
			if attached {FUNCTION [ANY, TUPLE, REAL_64]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_real_64, $Result)
			end
		end

	ce_class_item (a_args: TUPLE): CE_CLASS
		do
			check attached {FUNCTION [ANY, TUPLE, CE_CLASS]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
--			else
--				(create {EXCEPTIONS}).raise ("TODO (libffi bug)")
			end
		end

	pointer_item (a_args: TUPLE): CE_POINTER [ANY]
		local
			l_res: POINTER
		do
			if attached {FUNCTION [ANY, TUPLE, CE_POINTER [ANY]]} routine as tmp_function then
				Result := tmp_function.item (create_args (a_args))
			else
				ffi_execute (a_args, ffi_type_pointer, $l_res)
				create {CE_VOID_POINTER} Result.make_shared (l_res)
			end
		end

feature {NONE} -- create C function pointer from the agent

	create_c_function_pointer: POINTER
		local
			l_ret_eif_type: TYPE [detachable ANY]
			l_ret_ffi_type: POINTER
			l_old_state: BOOLEAN
		do
			if attached {PREDICATE [ANY, TUPLE]} routine then
				l_ret_ffi_type := ffi_type_integer_8
			elseif attached {FUNCTION [ANY, TUPLE, ANY]} routine as f then
				l_ret_eif_type := f.generating_type.generic_parameter_type (3)
				    if l_ret_eif_type = {INTEGER_8} then  l_ret_ffi_type := ffi_type_integer_8
				elseif l_ret_eif_type = {INTEGER_16} then l_ret_ffi_type := ffi_type_integer_16
				elseif l_ret_eif_type = {INTEGER_32} then l_ret_ffi_type := ffi_type_integer_32
				elseif l_ret_eif_type = {INTEGER_64} then l_ret_ffi_type := ffi_type_integer_64
				elseif l_ret_eif_type = {NATURAL_8} then  l_ret_ffi_type := ffi_type_natural_8
				elseif l_ret_eif_type = {NATURAL_16} then l_ret_ffi_type := ffi_type_natural_16
				elseif l_ret_eif_type = {NATURAL_32} then l_ret_ffi_type := ffi_type_natural_32
				elseif l_ret_eif_type = {NATURAL_64} then l_ret_ffi_type := ffi_type_natural_64
				elseif l_ret_eif_type = {REAL_32} then    l_ret_ffi_type := ffi_type_real_32
				elseif l_ret_eif_type = {REAL_64} then    l_ret_ffi_type := ffi_type_real_64
				elseif attached {TYPE [detachable CE_CLASS]} l_ret_eif_type as tmp then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					check attached {CE_CLASS} new_instance_of (tmp.type_id) as tmp_class then
						l_ret_ffi_type := ffi_type_struct (tmp_class.structure_size.to_natural_32)
					end
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
				elseif attached {TYPE [detachable CE_POINTER [ANY]]} l_ret_eif_type then
					l_ret_ffi_type := ffi_type_pointer
				else
					(create {EXCEPTIONS}).raise (generator + ".create_c_function_pointer: " + l_ret_eif_type)
				end
			else
				l_ret_ffi_type := ffi_type_void
			end
			Result := c_create_c_function_pointer ([$callback, Current], l_ret_ffi_type)
		end

	c_create_c_function_pointer (a_callback_data: TUPLE [POINTER, CE_ROUTINE]; a_ret_type: POINTER): POINTER
		external
			"C inline use <ffi.h>, <eif_setup.h>"
		alias
			"[
				void c_callback(ffi_cif *cif, void *ret, ffi_raw *args, void *callback_data)
				{
					#ifdef EIF_THREADS
						//EIF_INITIALIZE_AUX_THREAD; -> buggy before 7.2; workaround:
						GTCX;
						if(!eif_globals)
						{
							eif_thr_register(1);
							eif_set_thr_context ();
							initstk();
							#ifdef WORKBENCH
								xinitint();
							#endif
						}
	
						EIF_ENTER_EIFFEL;
						RTGC;
					#endif

					EIF_OBJECT t = (EIF_OBJECT)callback_data;
					void (*callback_func)() = eif_pointer_item(eif_access(t), 1);
					EIF_REFERENCE obj = eif_reference_item(eif_access(t), 2);
					(*callback_func)(obj, ret, args);

					#ifdef EIF_THREADS
						EIF_EXIT_EIFFEL;
						//eif_thr_exit();
					#endif
				}

				void (*created_fp)();
				ffi_cif *cif = malloc(sizeof(ffi_cif));
				ffi_raw_closure *closure = ffi_closure_alloc(sizeof(ffi_raw_closure), (void **)&created_fp);
				if(!closure) eif_panic("ffi_closure_alloc failed");
				if(ffi_prep_cif(cif, FFI_DEFAULT_ABI, 0, $a_ret_type, 0) != FFI_OK) eif_panic("ffi_prep_cif failed");
				if(ffi_prep_raw_closure_loc(closure, cif, c_callback, eif_adopt($a_callback_data), created_fp) != FFI_OK) eif_panic("ffi_prep_raw_closure_loc failed");
				return created_fp;
			]"
		end

	callback (a_ret, a_c_args: POINTER)
		local
			l_eif_args_call, l_eif_args_data: TUPLE -- different if vararg
			l_s: STRING_8
			l_c_args_managed: MANAGED_POINTER
			l_i, l_byte_index: INTEGER_32
			l_old_state: BOOLEAN
			l_ret_managed: MANAGED_POINTER
			l_ret_value: detachable ANY
		do
			check attached routine as tmp_routine then
				l_eif_args_call := tmp_routine.empty_operands
			end
			if attached {TUPLE [TUPLE]} l_eif_args_call then
				l_s := l_eif_args_call.generating_type.generic_parameter_type (1)
				l_s.remove_tail (1) -- remove closing "]"
				l_s.append (", CE_VA_LIST]")
				check attached {TUPLE} new_instance_of (dynamic_type_from_string (l_s)) as tmp then
					l_eif_args_data := tmp
				end
				l_eif_args_call.put_reference (l_eif_args_data, 1)
			else
				l_eif_args_data := l_eif_args_call
			end

			from
				create l_c_args_managed.share_from_pointer (a_c_args, {INTEGER_32}.Max_value)
				l_i := 1
			until
				l_i > l_eif_args_data.count
			loop
				inspect
					l_eif_args_data.item_code (l_i)
				when {TUPLE}.Boolean_code then
					l_eif_args_data.put_boolean (l_c_args_managed.read_boolean (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Character_8_code then
					l_eif_args_data.put_character_8 (l_c_args_managed.read_character (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Integer_8_code then
					l_eif_args_data.put_integer_8 (l_c_args_managed.read_integer_8 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Integer_16_code then
					l_eif_args_data.put_integer_16 (l_c_args_managed.read_integer_16 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Integer_32_code then
					l_eif_args_data.put_integer_32 (l_c_args_managed.read_integer_32 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes
				when {TUPLE}.Integer_64_code then
					l_eif_args_data.put_integer_64 (l_c_args_managed.read_integer_64 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_64_bytes
				when {TUPLE}.Natural_8_code then
					l_eif_args_data.put_natural_8 (l_c_args_managed.read_natural_8 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Natural_16_code then
					l_eif_args_data.put_natural_16 (l_c_args_managed.read_natural_16 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Integer_32_bytes -- promoted to `int' when passed through `...'
				when {TUPLE}.Natural_32_code then
					l_eif_args_data.put_natural_32 (l_c_args_managed.read_natural_32 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Natural_32_bytes
				when {TUPLE}.Natural_64_code then
					l_eif_args_data.put_natural_64 (l_c_args_managed.read_natural_64 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Natural_64_bytes
				when {TUPLE}.Real_32_code then
					l_eif_args_data.put_real_32 (l_c_args_managed.read_real_32 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Real_32_bytes
				when {TUPLE}.Real_64_code then
					l_eif_args_data.put_real_64 (l_c_args_managed.read_real_64 (l_byte_index), l_i)
					l_byte_index := l_byte_index + {PLATFORM}.Real_64_bytes
				else
					if l_i = l_eif_args_data.count and l_eif_args_call /= l_eif_args_data then
						l_eif_args_data.put_reference (create {CE_VA_LIST}.make_shared (l_c_args_managed.item + l_byte_index), l_i)
					elseif attached {TYPE [detachable CE_ROUTINE]} l_eif_args_data.generating_type.generic_parameter_type (l_i) as tmp then
						l_eif_args_data.put_reference (create {CE_ROUTINE}.make_shared (l_c_args_managed.read_pointer (l_byte_index)), l_i)
						l_byte_index := l_byte_index + {PLATFORM}.Pointer_bytes
					elseif attached {TYPE [detachable CE_POINTER [ANY]]} l_eif_args_data.generating_type.generic_parameter_type (l_i) as tmp then
						l_old_state := {ISE_RUNTIME}.check_assert (False)
						check
							attached {CE_POINTER [ANY]} new_instance_of (
									(create {CE_UTILITIES}).effective_type_id (True, tmp.generic_parameter_type (1))
								) as tmp_pointer
						then
							tmp_pointer.make_shared (l_c_args_managed.read_pointer (l_byte_index))
							l_eif_args_data.put_reference (tmp_pointer, l_i)
						end
						l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
						l_byte_index := l_byte_index + {PLATFORM}.Pointer_bytes
					elseif attached {TYPE [detachable CE_CLASS]} l_eif_args_data.generating_type.generic_parameter_type (l_i) as tmp then
						l_old_state := {ISE_RUNTIME}.check_assert (False)
						check attached {CE_CLASS} new_instance_of (tmp.type_id) as tmp_class then
							tmp_class.make_shared (l_c_args_managed.item + l_byte_index)
							l_eif_args_data.put_reference (tmp_class, l_i)
							l_byte_index := l_byte_index + tmp_class.structure_size
						end
						l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					elseif attached {TYPE [detachable CE_VA_LIST]} l_eif_args_data.generating_type.generic_parameter_type (l_i) then -- regular va_list argument, not vararg function
						l_eif_args_data.put_reference (create {CE_VA_LIST}.make_shared (l_c_args_managed.read_pointer (l_byte_index)), l_i)
						l_byte_index := l_byte_index + {PLATFORM}.Pointer_bytes
					else
						(create {EXCEPTIONS}).raise (generator + ".callback: " + l_eif_args_data.generating_type.generic_parameter_type (l_i))
					end
				end
				l_i := l_i + 1
			end

			check attached routine as tmp_routine then
				tmp_routine.call (l_eif_args_call)
			end

			if attached {FUNCTION [ANY, TUPLE, ANY]} routine as f then
				create l_ret_managed.share_from_pointer (a_ret, {INTEGER_32}.Max_value)
				l_ret_value := f.last_result
				    if attached {BOOLEAN} l_ret_value as tmp then    l_ret_managed.put_boolean (tmp, 0)
				elseif attached {INTEGER_8} l_ret_value as tmp then  l_ret_managed.put_integer_8 (tmp, 0)
				elseif attached {INTEGER_16} l_ret_value as tmp then l_ret_managed.put_integer_16 (tmp, 0)
				elseif attached {INTEGER_32} l_ret_value as tmp then l_ret_managed.put_integer_32 (tmp, 0)
				elseif attached {INTEGER_64} l_ret_value as tmp then l_ret_managed.put_integer_64 (tmp, 0)
				elseif attached {NATURAL_8} l_ret_value as tmp then  l_ret_managed.put_natural_8 (tmp, 0)
				elseif attached {NATURAL_16} l_ret_value as tmp then l_ret_managed.put_natural_16 (tmp, 0)
				elseif attached {NATURAL_32} l_ret_value as tmp then l_ret_managed.put_natural_32 (tmp, 0)
				elseif attached {NATURAL_64} l_ret_value as tmp then l_ret_managed.put_natural_64 (tmp, 0)
				elseif attached {REAL_32} l_ret_value as tmp then    l_ret_managed.put_real_32 (tmp, 0)
				elseif attached {REAL_64} l_ret_value as tmp then    l_ret_managed.put_real_64 (tmp, 0)
				elseif attached {CE_POINTER [ANY]} l_ret_value as tmp then
					l_ret_managed.put_pointer (tmp.c_pointer, 0)
				else
					if attached l_ret_value as tmp then
						(create {EXCEPTIONS}).raise (generator + ".create_c_function_pointer: " + tmp.generating_type)
					else
						(create {EXCEPTIONS}).raise (generator + ".create_c_function_pointer: Void")
					end
				end
			end
		end

	agent_c_pointers: HASH_TABLE [ROUTINE [ANY, TUPLE], POINTER]
		once ("PROCESS")
			create Result.make (128)
		end

feature {NONE} -- agent call helpers

	create_args (a_in_args_data: TUPLE): TUPLE
		require
			routine /= Void
			a_in_args_data /= Void
		local
			l_in_args_type: TYPE [detachable TUPLE]
			l_in_args_as_vararg: TUPLE [like a_in_args_data]

			l_out_args_data: TUPLE -- l_out_args_call <-> Result
			l_out_args_data_type: TYPE [detachable TUPLE]
			l_out_args_vararg_type: detachable TYPE [detachable TUPLE [TUPLE]]

			l_result: detachable TUPLE

			l_s: STRING
			l_i: INTEGER_32
			l_old_state: BOOLEAN
		do
			check
				attached routine as tmp_routine and then
				attached {TYPE [detachable TUPLE]} tmp_routine.generating_type.generic_parameter_type (2) as tmp
			then
				l_out_args_data_type := tmp
			end
			if attached {TYPE [detachable TUPLE [TUPLE]]} l_out_args_data_type as tmp then
				check attached {TYPE [detachable TUPLE]} tmp.generic_parameter_type (1) as tmp2 then
					l_out_args_data_type := tmp2
				end
				l_out_args_vararg_type := tmp
			end

			-- Try to directly use `a_in_args_data'. It most likely matches the out args type.
			if attached l_out_args_vararg_type as tmp then
				create l_in_args_as_vararg
				l_in_args_as_vararg.put_reference (a_in_args_data, 1)
				l_result := tmp.attempt (l_in_args_as_vararg)
			else
				l_result := l_out_args_data_type.attempt (a_in_args_data)
			end

			if attached l_result as tmp then
				Result := tmp
			else
				-- `a_in_args_data' doesn't match the out args type. Possible in C, catcall in Eiffel.
				check attached routine as tmp_routine then
					Result := tmp_routine.empty_operands
				end
				if attached l_out_args_vararg_type then
					l_s := l_out_args_data_type.name
					l_s.remove_tail (1) -- remove closing "]"
					from
						check attached {TYPE [detachable TUPLE]} a_in_args_data.generating_type as tmp then
							l_in_args_type := tmp
						end
						l_i := l_out_args_data_type.generic_parameter_count + 1
					until
						l_i > l_in_args_type.generic_parameter_count
					loop
						l_s.append (", ")
						l_s.append (l_in_args_type.generic_parameter_type (l_i))
						l_i := l_i + 1
					end
					l_s.append ("]")
					check attached {TUPLE} new_instance_of (dynamic_type_from_string (l_s)) as tmp then
						l_out_args_data := tmp
					end
					Result.put_reference (l_out_args_data, 1)
				else
					l_out_args_data := Result
				end

				from
					l_i := 1
				until
					l_i > l_out_args_data.count
				loop
					if
						not l_out_args_data.is_reference_item (l_i) or
						l_i > l_out_args_data_type.generic_parameter_count or
						attached {TYPE [detachable CE_VA_LIST]} l_out_args_data_type.generic_parameter_type (l_i)
					then
						l_out_args_data.put (a_in_args_data.item (l_i), l_i)
					elseif attached {TYPE [detachable CE_POINTER [ANY]]} l_out_args_data_type.generic_parameter_type (l_i) as tmp then
						l_old_state := {ISE_RUNTIME}.check_assert (False)
						check
							attached {CE_POINTER [ANY]} new_instance_of (
									(create {CE_UTILITIES}).effective_type_id (True, tmp.generic_parameter_type (1))
								) as tmp_out_pointer and
							attached {CE_POINTER [ANY]} a_in_args_data.reference_item (l_i) as tmp_in_pointer
						then
							tmp_out_pointer.make_cast (tmp_in_pointer)
							l_out_args_data.put_reference (tmp_out_pointer, l_i)
						end
						l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					else
						(create {EXCEPTIONS}).raise (generator + ".create_args: " + l_out_args_data_type.generic_parameter_type (l_i))
					end
					l_i := l_i + 1
				end
			end
		ensure
			Result /= Void
		end

feature {NONE} -- function pointer call helpers

	ffi_execute (a_args: TUPLE; a_ret_type, a_ret: POINTER)
		local
			l_arg_types, l_arg_values: POINTER
			l_i: INTEGER
			l_arg_pointers: ARRAY [POINTER]
			l_ffi_struct_types: TWO_WAY_LIST [POINTER]
		do
			if a_args.count > 0 then
				l_arg_types := l_arg_types.memory_alloc (a_args.count * {PLATFORM}.Pointer_bytes)
				l_arg_values := l_arg_values.memory_alloc (a_args.count * {PLATFORM}.Pointer_bytes)
			end
			from
				l_i := 1
				create l_arg_pointers.make_filled (default_pointer, 1, a_args.count)
				create l_ffi_struct_types.make
			until
				l_i > a_args.count
			loop
				inspect
					a_args.item_code (l_i)
				when {TUPLE}.Boolean_code then    ffi_set_boolean_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Integer_8_code then  ffi_set_integer_8_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Integer_16_code then ffi_set_integer_16_arg (l_arg_types, l_arg_values,l_i, a_args)
				when {TUPLE}.Integer_32_code then ffi_set_integer_32_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Integer_64_code then ffi_set_integer_64_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Natural_8_code then  ffi_set_natural_8_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Natural_16_code then ffi_set_natural_16_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Natural_32_code then ffi_set_natural_32_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Natural_64_code then ffi_set_natural_64_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Real_32_code then    ffi_set_real_32_arg (l_arg_types, l_arg_values, l_i, a_args)
				when {TUPLE}.Real_64_code then    ffi_set_real_64_arg (l_arg_types, l_arg_values, l_i, a_args)
				else
					if attached {CE_POINTER [ANY]} a_args.item (l_i) as tmp then
						l_arg_pointers[l_i] := tmp.c_pointer
						ffi_set_pointer_arg (l_arg_types, l_arg_values, l_i, l_arg_pointers.area.item_address (l_i - 1))
					elseif attached {CE_CLASS} a_args.item (l_i) as tmp then
						l_ffi_struct_types.extend (ffi_type_struct (tmp.structure_size.to_natural_32))
						ffi_set_struct_arg (l_arg_types, l_arg_values, l_i, l_ffi_struct_types.last, tmp.c_pointer)
					elseif attached {CE_VA_LIST} a_args.item (l_i) as tmp then
						l_arg_pointers[l_i] := tmp.c_pointer
						ffi_set_pointer_arg (l_arg_types, l_arg_values, l_i, l_arg_pointers.area.item_address (l_i - 1))
					elseif attached {REAL_96} a_args.item (l_i) as tmp then
						ffi_set_real_96_arg (l_arg_types, l_arg_values, l_i, tmp.c_pointer)
					else
						(create {EXCEPTIONS}).raise (generator + ".ffi_execute: " + a_args.generating_type.generic_parameter_type (l_i))
					end
				end
				l_i := l_i + 1
			end
			c_ffi_execute (a_args.count, a_ret_type, a_ret, c_pointer, l_arg_types, l_arg_values)
			from
				l_ffi_struct_types.start
			until
				l_ffi_struct_types.after
			loop
				l_ffi_struct_types.item.memory_free
				l_ffi_struct_types.forth
			end
		end

	c_ffi_execute (a_arg_count: INTEGER; a_ret_type, a_ret, a_routine, a_arg_types, a_arg_values: POINTER)
		external
			"C blocking inline use <ffi.h>"
		alias
			"[
				ffi_cif cif;
				if(ffi_prep_cif(&cif, FFI_DEFAULT_ABI, $a_arg_count, $a_ret_type, $a_arg_types) != FFI_OK) eif_panic("ffi_prep_cif failed!");
				ffi_call(&cif, $a_routine, $a_ret, $a_arg_values);
				if($a_arg_count > 0)
				{
					free($a_arg_types);
					free($a_arg_values);
				}
			]"
		end

feature {NONE} -- function pointer call helpers; `arg_types' and `arg_values' setters

	ffi_set_boolean_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_uint8;
				((void **)$a_arg_values)[$a_i - 1] = &eif_boolean_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_integer_8_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_sint8;
				((void **)$a_arg_values)[$a_i - 1] = &eif_integer_8_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_integer_16_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_sint16;
				((void **)$a_arg_values)[$a_i - 1] = &eif_integer_16_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_integer_32_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_sint32;
				((void **)$a_arg_values)[$a_i - 1] = &eif_integer_32_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_integer_64_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_sint64;
				((void **)$a_arg_values)[$a_i - 1] = &eif_integer_64_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_natural_8_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_uint8;
				((void **)$a_arg_values)[$a_i - 1] = &eif_natural_8_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_natural_16_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_uint16;
				((void **)$a_arg_values)[$a_i - 1] = &eif_natural_16_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_natural_32_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_uint32;
				((void **)$a_arg_values)[$a_i - 1] = &eif_natural_32_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_natural_64_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_uint64;
				((void **)$a_arg_values)[$a_i - 1] = &eif_natural_64_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_real_32_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_float;
				((void **)$a_arg_values)[$a_i - 1] = &eif_real_32_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_real_64_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_args: TUPLE)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_double;
				((void **)$a_arg_values)[$a_i - 1] = &eif_real_64_item(eif_access($a_args), $a_i);
			]"
		end

	ffi_set_real_96_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_p: POINTER)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_longdouble;
				((void **)$a_arg_values)[$a_i - 1] = $a_p;
			]"
		end

	ffi_set_pointer_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_p: POINTER)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = &ffi_type_pointer;
				((void **)$a_arg_values)[$a_i - 1] = $a_p;
			]"
		end

	ffi_set_struct_arg (a_arg_types, a_arg_values: POINTER; a_i: INTEGER; a_ffi_type_struct, a_p: POINTER)
		external
			"C inline use <ffi.h>"
		alias
			"[
				((ffi_type **)$a_arg_types)[$a_i - 1] = $a_ffi_type_struct;
				((void **)$a_arg_values)[$a_i - 1] = $a_p;
			]"
		end

feature {NONE} -- ffi types

	ffi_type_void: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_void;"
		end

	ffi_type_integer_8: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_sint8;"
		end

	ffi_type_integer_16: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_sint16;"
		end

	ffi_type_integer_32: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_sint32;"
		end

	ffi_type_integer_64: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_sint64;"
		end

	ffi_type_natural_8: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_uint8;"
		end

	ffi_type_natural_16: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_uint16;"
		end

	ffi_type_natural_32: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_uint32;"
		end

	ffi_type_natural_64: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_uint64;"
		end

	ffi_type_real_32: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_float;"
		end

	ffi_type_real_64: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_double;"
		end

	ffi_type_pointer: POINTER
		external
			"C inline use <ffi.h>"
		alias
			"return &ffi_type_pointer;"
		end

	ffi_type_struct (a_size: NATURAL_32): POINTER
		external
			"C inline use <ffi.h>"
		alias
			"[
				ffi_type *ffi_type_struct = malloc(sizeof(ffi_type));
				ffi_type_struct->size = $a_size;
				ffi_type_struct->alignment = ffi_type_sint.alignment;
				ffi_type_struct->type = FFI_TYPE_STRUCT;
				ffi_type_struct->elements = 0;
				return ffi_type_struct;
			]"
		end

feature {NONE} -- CE_POINTER features

	element_size: INTEGER_32
		do
			check False then
				-- Not yet implemented.
			end
		end

	item_from_pointer: ANY
		do
			check False then
				-- Not yet implemented.
			end
		end

	set_item_from_pointer (a_item: ANY)
		do
			check False then
				-- Not yet implemented.
			end
		end

	i_th_from_pointer (a_index: INTEGER_32): ANY
		do
			check False then
				-- Not yet implemented.
			end
		end

	set_i_th_from_pointer (a_item: ANY; a_index: INTEGER_32)
		do
			check False then
				-- Not yet implemented.
			end
		end

	plus_from_pointer (i: INTEGER_32): like Current
		do
			check False then
				-- Not yet implemented.
			end
		end

	minus_from_pointer (i: INTEGER_32): like Current
		do
			check False then
				-- Not yet implemented.
			end
		end

end
