class
	CE_STRUCTURE

create
	make_struct,
	make_union

feature {NONE}

	make_struct (a_fields: ARRAY [CE_STRUCTURE_FIELD])
		do
			make (True, a_fields)
		end

	make_union (a_fields: ARRAY [CE_STRUCTURE_FIELD])
		do
			make (False, a_fields)
		end

	make (a_is_struct: BOOLEAN; a_fields: ARRAY [CE_STRUCTURE_FIELD])
		do
			is_union := not a_is_struct
			is_struct := a_is_struct
			fields := a_fields.area

			create field_table.make (a_fields.count)
			across a_fields as c loop
				field_table[c.item.name] := c.item
				has_bitfields := has_bitfields or c.item.is_bitfield
			end
		end

feature {ANY}

	is_union, is_struct: BOOLEAN

	fields: SPECIAL [CE_STRUCTURE_FIELD]

	field_table: HASH_TABLE [CE_STRUCTURE_FIELD, STRING_8]

	size: INTEGER_32 assign set_size

	has_bitfields: BOOLEAN

	alignment: INTEGER_32 assign set_alignment

feature {ANY}

	set_superclasses (a_types: ARRAY [TYPE [detachable CE_CLASS]])
		require
			a_types /= Void and then not a_types.is_empty
		local
			l_internal: INTERNAL
			l_tmp_class: CE_CLASS
			l_most_fields: SPECIAL [CE_STRUCTURE_FIELD]
			l_fields: ARRAYED_LIST [CE_STRUCTURE_FIELD]
		do
			create l_internal
			check attached {CE_CLASS} l_internal.new_instance_of (a_types[1].type_id) as tmp then
				l_tmp_class := tmp
			end
			l_most_fields := l_tmp_class.structure_info.fields
			across a_types as c loop
				if
					c.target_index > 1 and then
					attached {CE_CLASS} l_internal.new_instance_of (c.item.type_id) as cur and then
					cur.structure_info.fields.count > l_most_fields.count
				then
					l_most_fields := cur.structure_info.fields
				end
			end

			create l_fields.make (l_most_fields.count + fields.count)
			across l_most_fields as c loop
				l_fields.extend (c.item.twin)
				has_bitfields := has_bitfields or c.item.is_bitfield
			end
			across fields as c loop
				l_fields.extend (c.item)
			end
			fields := l_fields.area

			create field_table.make (l_fields.count)
			across l_fields as c loop
				field_table[c.item.name] := c.item
			end
		end

	add_rename (a_superclass_name, a_subclass_name: STRING_8)
		require
			a_superclass_name /= Void and then not a_superclass_name.is_empty
			a_subclass_name /= Void and then not a_subclass_name.is_empty
			field_table.has (a_superclass_name)
			not field_table.has (a_subclass_name)
		do
			check attached field_table[a_superclass_name] as tmp then
				field_table[a_subclass_name] := tmp
				tmp.name := a_subclass_name
			end
		ensure
			field_table[a_subclass_name] = field_table[a_superclass_name]
			field_table.count = old field_table.count + 1
		end

	set_size (a_size: INTEGER_32)
		do
			size := a_size
		end

	set_alignment (a_alignment: INTEGER_32)
		do
			alignment := a_alignment
		end

invariant
	is_union /= is_struct
	fields /= Void
	field_table /= Void
	fields.count = field_table.count
	is_union implies across fields as c all c.item.is_bitfield or c.item.c_offset = 0 end

end
