class
	CE_UTILITIES

feature {ANY} -- string utilities

	ce_cstring (a_s: STRING_8): CE_ARRAY [INTEGER_8]
		require
			a_s /= Void
		do
			if attached static_string_storage[a_s] as tmp_result then
				Result := tmp_result
			else
				Result := ce_string (a_s)
				static_string_storage[a_s] := Result
			end
		ensure
			Result /= Void
		end

	ce_string (a_s: STRING_8): CE_ARRAY [INTEGER_8]
		require
			a_s /= Void
		local
			l_i: INTEGER_32
		do
			create {CE_INTEGER_8_ARRAY} Result.make (<<a_s.count + 1>>)
			from
				l_i := 1
			until
				l_i > a_s.count
			loop
				Result[l_i - 1] := a_s[l_i].code.to_integer_8
				l_i := l_i + 1
			end
		ensure
			Result /= Void
		end

	eif_string (a_p: CE_POINTER [INTEGER_8]): STRING
		require
			a_p /= Void and then not a_p.is_void
		do
			create Result.make_from_c (a_p.c_pointer)
		ensure
			Result /= Void
		end

feature {ANY} -- ce-type utilities

	Ce_type_boolean,
	Ce_type_integer_8,
	Ce_type_integer_16,
	Ce_type_integer_32,
	Ce_type_integer_64,
	Ce_type_natural_8,
	Ce_type_natural_16,
	Ce_type_natural_32,
	Ce_type_natural_64,
	Ce_type_real_32,
	Ce_type_real_64,
	Ce_type_real_96,
	Ce_type_void,
	Ce_type_va_list,
	Ce_type_pointer,
	Ce_type_array,
	Ce_type_class: INTEGER_32 = unique

	ce_type (a_eif_type: TYPE [detachable ANY]): INTEGER_32
		require
			a_eif_type /= Void
		do
			    if a_eif_type = {BOOLEAN}    then Result := Ce_type_boolean
			elseif a_eif_type = {INTEGER_8}  then Result := Ce_type_integer_8
			elseif a_eif_type = {INTEGER_16} then Result := Ce_type_integer_16
			elseif a_eif_type = {INTEGER_32} then Result := Ce_type_integer_32
			elseif a_eif_type = {INTEGER_64} then Result := Ce_type_integer_64
			elseif a_eif_type = {NATURAL_8}  then Result := Ce_type_natural_8
			elseif a_eif_type = {NATURAL_16} then Result := Ce_type_natural_16
			elseif a_eif_type = {NATURAL_32} then Result := Ce_type_natural_32
			elseif a_eif_type = {NATURAL_64} then Result := Ce_type_natural_64
			elseif a_eif_type = {REAL_32}    then Result := Ce_type_real_32
			elseif a_eif_type = {REAL_64}    then Result := Ce_type_real_64
			elseif a_eif_type = {REAL_96}    then Result := Ce_type_real_96
			elseif a_eif_type = {ANY}        then Result := Ce_type_void
			elseif a_eif_type = {CE_VA_LIST} then Result := Ce_type_va_list
			elseif attached {TYPE [detachable CE_ARRAY [ANY]]} a_eif_type   then Result := Ce_type_array
			elseif attached {TYPE [detachable CE_POINTER [ANY]]} a_eif_type then Result := Ce_type_pointer
			elseif attached {TYPE [detachable CE_CLASS]} a_eif_type         then Result := Ce_type_class
			else
				(create {EXCEPTIONS}).raise (a_eif_type.name)
			end
		end

	effective_type_id (a_is_pointer: BOOLEAN; a_generic_type: TYPE [detachable ANY]): INTEGER_32
			-- `INTEGER_16' -> `CE_INTEGER_16_POINTER'
			-- `CE_POINTER [INTEGER_16]' -> `CE_POINTER_POINTER [INTEGER_16]'
			-- `CE_ARRAY [INTEGER_16]' -> `CE_ARRAY_POINTER [INTEGER_16]'
			-- `[PERSON]' -> `CE_CLASS_POINTER [PERSON]'
		do
			if a_is_pointer then
				    if a_generic_type = {ANY}        then Result := ({CE_VOID_POINTER}).type_id
				elseif a_generic_type = {BOOLEAN}    then Result := ({CE_BOOLEAN_POINTER}).type_id
				elseif a_generic_type = {INTEGER_8}  then Result := ({CE_INTEGER_8_POINTER}).type_id
				elseif a_generic_type = {INTEGER_16} then Result := ({CE_INTEGER_16_POINTER}).type_id
				elseif a_generic_type = {INTEGER_32} then Result := ({CE_INTEGER_32_POINTER}).type_id
				elseif a_generic_type = {INTEGER_64} then Result := ({CE_INTEGER_64_POINTER}).type_id
				elseif a_generic_type = {NATURAL_8}  then Result := ({CE_NATURAL_8_POINTER}).type_id
				elseif a_generic_type = {NATURAL_16} then Result := ({CE_NATURAL_16_POINTER}).type_id
				elseif a_generic_type = {NATURAL_32} then Result := ({CE_NATURAL_32_POINTER}).type_id
				elseif a_generic_type = {NATURAL_64} then Result := ({CE_NATURAL_64_POINTER}).type_id
				elseif a_generic_type = {REAL_32}    then Result := ({CE_REAL_32_POINTER}).type_id
				elseif a_generic_type = {REAL_64}    then Result := ({CE_REAL_64_POINTER}).type_id
				elseif a_generic_type = {REAL_96}    then Result := ({CE_REAL_96_POINTER}).type_id
				elseif a_generic_type = {CE_ROUTINE} then Result := ({CE_ROUTINE_POINTER}).type_id
				elseif a_generic_type = {CE_VA_LIST} then Result := ({CE_VA_LIST_POINTER}).type_id
				elseif attached {TYPE [detachable CE_ARRAY [ANY]]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_ARRAY_POINTER [" + a_generic_type.generic_parameter_type (1).name + "]")
				elseif attached {TYPE [detachable CE_POINTER [ANY]]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_POINTER_POINTER [" + a_generic_type.generic_parameter_type (1).name + "]")
				elseif attached {TYPE [detachable CE_CLASS]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_CLASS_POINTER [" + a_generic_type.name + "]")
				else
					(create {EXCEPTIONS}).raise ("effective_type_id: " + a_generic_type)
				end
			else
				    if a_generic_type = {BOOLEAN}    then Result := ({CE_BOOLEAN_ARRAY}).type_id
				elseif a_generic_type = {INTEGER_8}  then Result := ({CE_INTEGER_8_ARRAY}).type_id
				elseif a_generic_type = {INTEGER_16} then Result := ({CE_INTEGER_16_ARRAY}).type_id
				elseif a_generic_type = {INTEGER_32} then Result := ({CE_INTEGER_32_ARRAY}).type_id
				elseif a_generic_type = {INTEGER_64} then Result := ({CE_INTEGER_64_ARRAY}).type_id
				elseif a_generic_type = {NATURAL_8}  then Result := ({CE_NATURAL_8_ARRAY}).type_id
				elseif a_generic_type = {NATURAL_16} then Result := ({CE_NATURAL_16_ARRAY}).type_id
				elseif a_generic_type = {NATURAL_32} then Result := ({CE_NATURAL_32_ARRAY}).type_id
				elseif a_generic_type = {NATURAL_64} then Result := ({CE_NATURAL_64_ARRAY}).type_id
				elseif a_generic_type = {REAL_32}    then Result := ({CE_REAL_32_ARRAY}).type_id
				elseif a_generic_type = {REAL_64}    then Result := ({CE_REAL_64_ARRAY}).type_id
				elseif a_generic_type = {REAL_96}    then Result := ({CE_REAL_96_ARRAY}).type_id
				elseif a_generic_type = {CE_ROUTINE} then Result := ({CE_ROUTINE_ARRAY}).type_id
				elseif a_generic_type = {CE_VA_LIST} then Result := ({CE_VA_LIST_ARRAY}).type_id
				elseif attached {TYPE [detachable CE_ARRAY [ANY]]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_ARRAY_ARRAY [" + a_generic_type.generic_parameter_type (1).name + "]")
				elseif attached {TYPE [detachable CE_POINTER [ANY]]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_POINTER_ARRAY [" + a_generic_type.generic_parameter_type (1).name + "]")
				elseif attached {TYPE [detachable CE_CLASS]} a_generic_type then
					Result := (create {INTERNAL}).dynamic_type_from_string ("CE_CLASS_ARRAY [" + a_generic_type.name + "]")
				else
					(create {EXCEPTIONS}).raise ("effective_type_id: " + a_generic_type)
					-- make sure all implementation classes are compiled/present:
					(
						attached {CE_ARRAY_ARRAY [ANY]} Current or
						attached {CE_POINTER_ARRAY [ANY]} Current or
						attached {CE_CLASS_ARRAY [CE_CLASS]} Current or
						attached {CE_ARRAY_POINTER [ANY]} Current or -- independent from {CE_ARRAY_ARRAY} above
						-- why is the following needed all the sudden?
						attached {CE_ARRAY_ARRAY [INTEGER_32]} Current or
						attached {CE_ARRAY_ARRAY [REAL_32]} Current
					).do_nothing
				end
			end
		ensure
			Result > -1
		end

feature {ANY} -- alignment utilities

	ce_alignof (a_eif_type: TYPE [detachable ANY]): NATURAL_32
		require
			a_eif_type /= Void
		local
			l_internal: INTERNAL
		do
			inspect ce_type (a_eif_type)
			when Ce_type_void, Ce_type_boolean, Ce_type_integer_8, Ce_type_natural_8 then
				Result := 1
			when Ce_type_integer_16, Ce_type_natural_16 then
				Result := 2
			when
				Ce_type_integer_32, Ce_type_integer_64,
				Ce_type_natural_32, Ce_type_natural_64,
				Ce_type_real_32, Ce_type_real_64, Ce_type_real_96,
				Ce_type_pointer, Ce_type_va_list
			then
				Result := 4
			when Ce_type_array then
				if a_eif_type.generic_parameter_count > 0 then
					Result := ce_alignof (a_eif_type.generic_parameter_type (1))
				else
					    if a_eif_type = {CE_BOOLEAN_ARRAY}    then Result := ce_alignof ({BOOLEAN})
					elseif a_eif_type = {CE_INTEGER_8_ARRAY}  then Result := ce_alignof ({INTEGER_8})
					elseif a_eif_type = {CE_INTEGER_16_ARRAY} then Result := ce_alignof ({INTEGER_16})
					elseif a_eif_type = {CE_INTEGER_32_ARRAY} then Result := ce_alignof ({INTEGER_32})
					elseif a_eif_type = {CE_INTEGER_64_ARRAY} then Result := ce_alignof ({INTEGER_64})
					elseif a_eif_type = {CE_NATURAL_8_ARRAY}  then Result := ce_alignof ({NATURAL_8})
					elseif a_eif_type = {CE_NATURAL_16_ARRAY} then Result := ce_alignof ({NATURAL_16})
					elseif a_eif_type = {CE_NATURAL_32_ARRAY} then Result := ce_alignof ({NATURAL_32})
					elseif a_eif_type = {CE_NATURAL_64_ARRAY} then Result := ce_alignof ({NATURAL_64})
					elseif a_eif_type = {CE_REAL_32_ARRAY}    then Result := ce_alignof ({REAL_32})
					elseif a_eif_type = {CE_REAL_64_ARRAY}    then Result := ce_alignof ({REAL_64})
					elseif a_eif_type = {CE_REAL_96_ARRAY}    then Result := ce_alignof ({REAL_96})
					elseif a_eif_type = {CE_ROUTINE_ARRAY}    then Result := ce_alignof ({CE_ROUTINE})
					else
						(create {EXCEPTIONS}).raise (a_eif_type)
					end
				end
			when Ce_type_class then
				create l_internal
				check attached {CE_CLASS} l_internal.new_instance_of (a_eif_type.type_id) as tmp_class then
					if tmp_class.structure_info.alignment /= 0 then
						Result := tmp_class.structure_info.alignment.to_natural_32
					elseif
						attached tmp_class.structure_info.fields as f and then
						f.count > 0
					then
						if f[0].alignment /= 0 then
							Result := f[0].alignment.to_natural_32
						else
							Result := ce_alignof (l_internal.type_of_type (f[0].eif_type))
						end
					else
						Result := 1
					end
				end
			end
		ensure
			Result = 1 or Result = 2 or Result = 4
		end

feature {ANY} -- REAL_96 utilities

	Real_96_bytes: INTEGER_32 = 12

	real_96_from_integer_8 (a_value: INTEGER_8): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_integer_8 (a_value)
		end

	real_96_from_integer_16 (a_value: INTEGER_16): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_integer_16 (a_value)
		end

	real_96_from_integer_32 (a_value: INTEGER_32): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_integer_32 (a_value)
		end

	real_96_from_integer_64 (a_value: INTEGER_64): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_integer_64 (a_value)
		end

	real_96_from_natural_8 (a_value: NATURAL_8): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_natural_8 (a_value)
		end

	real_96_from_natural_16 (a_value: NATURAL_16): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_natural_16 (a_value)
		end

	real_96_from_natural_32 (a_value: NATURAL_32): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_natural_32 (a_value)
		end

	real_96_from_natural_64 (a_value: NATURAL_64): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_natural_64 (a_value)
		end

	real_96_from_real_32 (a_value: REAL_32): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_real_32 (a_value)
		end

	real_96_from_real_64 (a_value: REAL_64): REAL_96
		do
			a_value.do_nothing -- bug 17924
			Result.set_from_real_64 (a_value)
		end

feature {ANY} -- hex floating point utilities

	hex_to_real_32 (a_str: STRING_8): REAL_32
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_str)
			Result := c_hex_to_real_32 (l_c_str.item)
		end

	hex_to_real_64 (a_str: STRING_8): REAL_64
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_str)
			Result := c_hex_to_real_64 (l_c_str.item)
		end

	hex_to_real_96 (a_str: STRING_8): REAL_96
		local
			l_c_str: C_STRING
		do
			create l_c_str.make (a_str)
			c_hex_to_real_96 (Result.c_pointer, l_c_str.item)
		end

feature {NONE}

	static_string_storage: HASH_TABLE [CE_ARRAY [INTEGER_8], STRING_8]
		once ("PROCESS")
			create Result.make (16)
		end

	c_hex_to_real_32 (a_str: POINTER): REAL_32
		external
			"C inline use <stdlib.h>"
		alias
			"return strtof($a_str, 0);"
		end

	c_hex_to_real_64 (a_str: POINTER): REAL_64
		external
			"C inline use <stdlib.h>"
		alias
			"return strtod($a_str, 0);"
		end

	c_hex_to_real_96 (a_res, a_str: POINTER)
		external
			"C inline use <stdlib.h>"
		alias
			"*(long double *)$a_res = strtold($a_str, 0);"
		end

end
