class
	CE_ARRAY_ARRAY [G]

inherit
	CE_ARRAY [CE_ARRAY [G]]
		redefine
			make,
			make_shared,
			make_shared_from_pointer,
			make_initialized,
			copy
		end
	INTERNAL
		rename
			is_equal as is_equal_from_any
		undefine
			is_equal_from_any
		redefine
			copy
		end

create
	make,
	make_shared,
	make_initialized

feature {NONE}

	make (a_dimensions: ARRAY [INTEGER_32])
		require else
			a_dimensions /= Void
			a_dimensions.count >= 1
			a_dimensions[1] >= 0
		do
			make_shared (c_pointer, a_dimensions)
			c_pointer := c_pointer.memory_calloc (total_count, element_size)
			is_allocated_memory := True
		end

	make_shared (a_c_pointer: POINTER; a_dimensions: ARRAY [INTEGER_32])
		require else
			a_dimensions /= Void
			a_dimensions.count >= 1
			a_dimensions[1] >= 0
		local
			l_i: INTEGER_32
			l_total_count: INTEGER_32
			l_subarray_dimensions: ARRAY [INTEGER_32]
		do
			init_element_info

			from
				l_i := 2
				l_total_count := a_dimensions[1]
				create l_subarray_dimensions.make_filled (0, 1, a_dimensions.count - 1)
			until
				l_i > a_dimensions.count
			loop
				l_total_count := l_total_count * a_dimensions[l_i]
				l_subarray_dimensions[l_i - 1] := a_dimensions[l_i]
				l_i := l_i + 1
			end

			byte_count := a_dimensions[1] * element_size
			c_pointer := a_c_pointer
			total_count := l_total_count

			subarray_dimensions := l_subarray_dimensions
			create subarray_cache.make_filled (Void, 0, a_dimensions[1] - 1)
		end

	make_initialized (a_data: ARRAY [CE_ARRAY [G]])
		local
			l_dimensions: ARRAYED_LIST [INTEGER_32]
		do
			create l_dimensions.make (16)
			l_dimensions.extend (a_data.count)
			l_dimensions.extend (a_data[1].count)
			if attached {CE_ARRAY_ARRAY [ANY]} a_data[1] as tmp then
				across tmp.subarray_dimensions as c loop
					l_dimensions.extend (c.item)
				end
			end
			make (l_dimensions.to_array)

			across a_data as c loop
				i_th (c.target_index - 1).c_pointer.memory_copy (
						c.item.c_pointer,
						c.item.total_count * c.item.element_size
					)
			end
		end

feature {ANY}

	element_size: INTEGER_32

feature {ANY}

	i_th alias "[]" (a_index: INTEGER_32): CE_ARRAY [G] assign set_i_th
			-- Current[0]
		local
			l_old_state: BOOLEAN
		do
			if attached subarray_cache[a_index] as tmp_result then
				Result := tmp_result
			else
				l_old_state := {ISE_RUNTIME}.check_assert (False)
				check attached {CE_ARRAY [G]} new_instance_of (element_type_id) as tmp_result then
					tmp_result.make_shared (c_pointer + (a_index * (total_count // count) * element_size), subarray_dimensions)
					Result := tmp_result
				end
				l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
				subarray_cache[a_index] := Result
			end
		end

	copy (a_other: like Current)
		local
			l_dimensions: ARRAY [INTEGER_32]
		do
			create l_dimensions.make_filled (0, 1, a_other.subarray_dimensions.count + 1)
			l_dimensions[1] := a_other.count
			l_dimensions.subcopy (a_other.subarray_dimensions, 1, a_other.subarray_dimensions.count, 2)
			make (l_dimensions)
			c_pointer.memory_copy (a_other.c_pointer, total_count * element_size)
		end

	plus alias "+" (i: INTEGER_32): CE_ARRAY_ARRAY [G]
		do
			check False then
				-- Not yet implemented.
			end
		end

feature {CE_ARRAY_ARRAY}

	subarray_dimensions: ARRAY [INTEGER_32]

	subarray_cache: ARRAY [detachable CE_ARRAY [G]]

feature {NONE}

	element_info_cache: SPECIAL [INTEGER_32]
		once ("PROCESS")
			create Result.make_filled (-1, 10_000)
		end

	init_element_info
		local
			l_type_id: INTEGER_32
			l_old_state: BOOLEAN
		do
			l_type_id := ({G}).type_id
			element_type_id := element_info_cache[2 * l_type_id]
			element_size := element_info_cache[2 * l_type_id + 1]

			if element_type_id = -1 then
				element_type_id := (create {CE_UTILITIES}).effective_type_id (False, {G})

				l_old_state := {ISE_RUNTIME}.check_assert (False)
				check attached {CE_ARRAY [ANY]} new_instance_of (element_type_id) as tmp_array then
					element_size := tmp_array.element_size
					if element_size = 0 then -- CE_ARRAY_ARRAY
						tmp_array.make_shared (default_pointer + 1, <<0>>)
						element_size := tmp_array.element_size
					end
				end
				l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)

				element_info_cache[2 * l_type_id] := element_type_id
				element_info_cache[2 * l_type_id + 1] := element_size
			end
		end

	element_type_id: INTEGER_32

feature {NONE}

	make_shared_from_pointer (a_c_pointer: POINTER)
		do
			(create {EXCEPTIONS}).raise ("Not possible for multi-dimension arrays.")
		end

	set_i_th (a_item: CE_ARRAY [G]; a_index: INTEGER_32)
		do
			(create {EXCEPTIONS}).raise ("Not possible for multi-dimension arrays.")
		end

invariant
	subarray_dimensions /= Void and then not subarray_dimensions.is_empty
	subarray_cache /= Void

end
